/*
 * mc.c
 */

#include "multilevel.h"
#include "mc.h"
#include "mesh.h"
#include "voxels.h"
#include <math.h>

/******************************************************************/
/* Translation tables for use with the below MC tables.           */

/*
  My corner/edge map has the advantage that the number of each corner
  is obtained by corner = x | (y << 1) | (z << 2).

                _4____________4_____________5
               /|                           /
              / |                          /|
             /  |                         / |
            7   |                        /  |
           /    |                       /5  |
          /     |                      /    |
         /      8                     /     9
        /       |                    /      |
      6/________|______6____________/7      |
       |        |                   |       |
       |        |                   |       |
       |        |                   |       |
       |        |0____________0_____|_______|1
      11       /                    |      /
       |      /                    10     /
       |     /                      |    /
       |    /3                      |   /1
       |   /                        |  /
       |  /                         | /
       | /                          |/
       |/2____________2_____________|3

 */

static int corner2vertex[8] = { 0, 1, 3, 2, 4, 5, 7, 6 };
static int edge2corners[12][2] = {
  { 0, 1 }, { 1, 3 }, { 2, 3 }, { 0, 2 },
  { 4, 5 }, { 5, 7 }, { 6, 7 }, { 4, 6 },
  { 0, 4 }, { 1, 5 }, { 3, 7 }, { 2, 6 }
};


/******************************************************************/
/* Taken from http://www.erc.msstate.edu/~chupa/f97/vis/lab1/mc.h */

/*
  these tables for computing the Marching Cubes algorithm
  are from http://www.mhri.edu.au/~pdb/modelling/polygonise/
  by Paul Bourke, based on code by Cory Gene Bloyd.

  The indexing of vertices and edges in a cube are defined
  as: 
  
                _4____________4_____________5
               /|                           /
              / |                          /|
             /  |                         / |
            7   |                        /  |
           /    |                       /5  |
          /     |                      /    |
         /      8                     /     9
        /       |                    /      |
      7/________|______6____________/6      |
       |        |                   |       |
       |        |                   |       |
       |        |                   |       |
       |        |0____________0_____|_______|1
      11       /                    |      /
       |      /                    10     /
       |     /                      |    /
       |    /3                      |   /1
       |   /                        |  /
       |  /                         | /
       | /                          |/
       |/3____________2_____________|2

 


For purposes of calculating vertices along the edges and the
triangulations created, there are 15 distinct cases, with
upper limits of
  12 edge intersections
  5 triangles created per cell
*/

static int edgeTable[256]={
  0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
  0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
  0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
  0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
  0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
  0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
  0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
  0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
  0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
  0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
  0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
  0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
  0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
  0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
  0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
  0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
  0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
  0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
  0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
  0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
  0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
  0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
  0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
  0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
  0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
  0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
  0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
  0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
  0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
  0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
  0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
  0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0   };


static int triTable[256][16] =
{
 {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
 {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
 {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
 {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
 {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
 {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
 {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
 {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
 {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
 {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
 {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
 {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
 {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
 {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
 {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
 {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
 {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
 {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
 {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
 {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
 {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
 {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
 {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
 {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
 {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
 {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
 {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
 {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
 {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
 {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
 {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
 {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
 {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
 {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
 {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
 {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
 {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
 {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
 {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
 {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
 {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
 {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
 {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
 {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
 {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
 {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
 {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
 {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
 {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
 {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
 {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
 {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
 {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
 {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
 {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
 {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
 {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
 {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
 {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
 {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
 {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
 {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
 {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
 {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
 {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
 {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
 {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
 {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
 {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
 {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
 {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
 {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
 {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
 {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
 {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
 {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
 {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
 {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
 {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
 {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
 {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
 {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
 {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
 {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
 {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
 {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
 {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
 {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
 {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
 {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
 {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
 {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
 {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
 {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
 {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
 {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
 {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
 {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
 {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
 {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
 {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
 {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
 {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
 {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
 {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
 {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
 {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
 {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
 {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
 {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
 {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
 {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
 {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
 {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
 {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
 {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
 {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
 {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
 {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
 {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
 {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
 {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
 {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
 {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
 {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
 {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
 {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
 {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
 {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
 {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
 {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
 {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
 {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
 {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
 {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
 {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
 {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
 {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
 {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
 {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
 {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
 {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
 {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
 {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
 {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
 {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
 {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
 {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
 {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
 {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
 {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
 {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
 {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
 {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
 {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
 {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
 {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
 {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
 {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
 {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
 {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
 {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
 {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
 {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
 {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
 {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
 {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
 {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
 {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
 {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
 {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
 {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
 {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
 {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
 {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
 {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
 {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
 {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
 {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
 {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
 {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
};




/******************************************************************************/
/* Adaptive marching cubes algorithm. */

static int
binary_search_x(int x, int y, int z, int x2)
{
  assert(x2 - x >= 1);
  assert(x >= 0);
  assert(x2 >= 0);
  assert(y >= 0);
  assert(z >= 0);
  assert(x <= DIMX);
  assert(x2 <= DIMX+1);
  assert(y <= DIMY);
  assert(z <= DIMZ);

  if (x2 - x == 1)
    return x;
  if ((voxels[(x + x2) / 2][y][z] > 127) != (voxels[x][y][z] > 127)) {
    return binary_search_x(x, y, z, (x + x2) / 2);
  } else {
    return binary_search_x((x + x2) / 2, y, z, x2);
  }
} 

static int
binary_search_y(int x, int y, int z, int y2)
{
  assert(y2 - y >= 1);
  assert(x >= 0);
  assert(y >= 0);
  assert(z >= 0);
  assert(y2 >= 0);
  assert(x <= DIMX);
  assert(y <= DIMY);
  assert(z <= DIMZ);
  assert(y2 <= DIMY+1);

  if (y2 - y == 1)
    return y;
  if ((voxels[x][(y + y2) / 2][z] > 127) != (voxels[x][y][z] > 127)) {
    return binary_search_y(x, y, z, (y + y2) / 2);
  } else {
    return binary_search_y(x, (y + y2) / 2, z, y2);
  }
} 

static int
binary_search_z(int x, int y, int z, int z2)
{
  assert(z2 - z >= 1);
  assert(x >= 0);
  assert(y >= 0);
  assert(z >= 0);
  assert(z2 >= 0);
  assert(x <= DIMX);
  assert(y <= DIMY);
  assert(z <= DIMZ);
  assert(z2 <= DIMZ+1);

  if (z2 - z == 1)
    return z;
  if ((voxels[x][y][(z + z2) / 2] > 127) != (voxels[x][y][z] > 127)) {
    return binary_search_z(x, y, z, (z + z2) / 2);
  } else {
    return binary_search_z(x, y, (z + z2) / 2, z2);
  }
} 



static int
march_cube(int  x, int  y, int  z,
           int sx, int sy, int sz,
           Vertex *t[], int *nt,
           int setadj)
{
  uint8 index = 0;
  int i, j, k;

  for (i = 0; i < 2; i++) {
    for (j = 0; j < 2; j++) {
      for (k = 0; k < 2; k++) {
        if (voxels[x+i*sx][y+j*sy][z+k*sz] > 127)
          index |= 1 << corner2vertex[i | (j << 1) | (k << 2)];
      }
    }
  }

  for (i = 0; triTable[index][i] >= 0; i += 3) {
    t[i] = t[i+1] = t[i+2] = 0;
    for (j = 0; j < 3; j++) {
      int corner0, corner1, diff;
      int ex, ey, ez;
      float xx, yy, zz;
      float nx, ny, nz;
      float delta;
      int tag;
      float v[3], n[3];

      corner0 = edge2corners[triTable[index][i+j]][0];
      corner1 = edge2corners[triTable[index][i+j]][1];
      diff = corner0 ^ corner1;

      ex = x + (corner0 & 1 ? 1 : 0)*sx;
      ey = y + (corner0 & 2 ? 1 : 0)*sy;
      ez = z + (corner0 & 4 ? 1 : 0)*sz;

      xx = (float)ex / (DIMX+1);
      yy = (float)ey / (DIMY+1);
      zz = (float)ez / (DIMZ+1);

      if (diff & 1) {
        ex = binary_search_x(ex, ey, ez, ex + sx);
      } else if (diff & 2) {
        ey = binary_search_y(ex, ey, ez, ey + sy);
      } else if (diff & 4) {
        ez = binary_search_z(ex, ey, ez, ez + sz);
      } else {
        g_print("Er, uh, oops, diff == %01X\n", diff);
      }

      /* This definition of tag is bogus, it'll fail for DIMX/Y/Z > 1023, but oh well. */
      tag = ((diff & 1 ? 1 : diff & 2 ? 2 : diff & 4 ? 3 : 0) << 30)
        | (ez << 20) | (ey << 10) | (ex);
      if ((t[i+j] = get_vertex(tag)) != 0) {

#if 0
        if (diff & 1) {
          t[i+j]->cr = 1.0;
          t[i+j]->cg = 0.0;
          t[i+j]->cb = 0.0;
        } else if (diff & 2) {
          t[i+j]->cr = 0.0;
          t[i+j]->cg = 1.0;
          t[i+j]->cb = 0.0;
        } else if (diff & 4) {
          t[i+j]->cr = 0.0;
          t[i+j]->cg = 0.0;
          t[i+j]->cb = 1.0;
        } else {
          g_print("Er, uh, oops, diff == %01X\n", diff);
        }
#endif

        if (setadj) {
          /* Set the adjacency information and continue without creating a new vertex. */
          if (diff & 1) {
            t[i+j]->adj |= 1 << ((corner0 & 2 ? 1 : 0) | (corner0 & 4 ? 2 : 0));
          } else if (diff & 2) {
            t[i+j]->adj |= 1 << ((corner0 & 4 ? 1 : 0) | (corner0 & 1 ? 2 : 0));
          } else if (diff & 4) {
            t[i+j]->adj |= 1 << ((corner0 & 1 ? 1 : 0) | (corner0 & 2 ? 2 : 0));
          } else {
            g_print("Er, uh, oops, diff == %01X\n", diff);
          }
        }

        continue;
      }

      if (diff & 1) {
        delta = (127.5 - voxels[ex][ey][ez])/(voxels[ex+1][ey][ez] - voxels[ex][ey][ez]);
        xx = ((float)ex + delta)/(DIMX+1);
      } else if (diff & 2) {
        delta = (127.5 - voxels[ex][ey][ez])/(voxels[ex][ey+1][ez] - voxels[ex][ey][ez]);
        yy = ((float)ey + delta)/(DIMY+1);
      } else if (diff & 4) {
        delta = (127.5 - voxels[ex][ey][ez])/(voxels[ex][ey][ez+1] - voxels[ex][ey][ez]);
        zz = ((float)ez + delta)/(DIMZ+1);
      } else {
        g_print("Er, uh, oops, diff == %01X\n", diff);
      }

      voxel_normal(xx, yy, zz, &nx, &ny, &nz);

      v[0] = xx; v[1] = yy; v[2] = zz;
      n[0] = nx; n[1] = ny; n[2] = nz;
      t[i+j] = add_vertex(tag, v, n);

      if (setadj) {
        if (diff & 1) {
          t[i+j]->adj |= 1 << ((corner0 & 2 ? 1 : 0) | (corner0 & 4 ? 2 : 0));
        } else if (diff & 2) {
          t[i+j]->adj |= 1 << ((corner0 & 4 ? 1 : 0) | (corner0 & 1 ? 2 : 0));
        } else if (diff & 4) {
          t[i+j]->adj |= 1 << ((corner0 & 1 ? 1 : 0) | (corner0 & 2 ? 2 : 0));
        } else {
          g_print("Er, uh, oops, diff == %01X\n", diff);
        }
      }

    }

  }
  *nt = i/3;

  return 0;
}


#define ADAPTIVE_ERROR 40.0
static float
adaptive_error(int  x, int  y, int  z,
               int sx, int sy, int sz,
               Vertex *t[], int nt)
{
  int i, j;
  float centroid[3];
  float error = 0.0;

  for (i = 0; i < nt; i++) {

    centroid[0] = centroid[1] = centroid[2] = 0.0;
    for (j = 0; j < 3; j++) {
      centroid[0] += t[3*i+j]->x;
      centroid[1] += t[3*i+j]->y;
      centroid[2] += t[3*i+j]->z;
    }
    centroid[0] /= 3.0;
    centroid[1] /= 3.0;
    centroid[2] /= 3.0;

    error += fabs(voxel_eval(centroid) - 127.5);
  }
#if 0
  if (nt > 0)
    g_print("adaptive_error %f at (%i,%i,%i)-(%i,%i,%i), %i tris\n",
            error, x, y, z, x+sx, y+sy, z+sz, nt);
#endif
  return nt > 0 ? error / nt : 0.0;
}


int
adaptive_march_cube(int maxdepth,
                    int  x, int  y, int  z,
                    int sx, int sy, int sz)
{
  static Vertex *t[15];
  int nt, i;

  assert(x >= 0);
  assert(y >= 0);
  assert(z >= 0);
  assert(x <= DIMX);
  assert(y <= DIMY);
  assert(z <= DIMZ);

  march_cube(x, y, z, sx, sy, sz, t, &nt, 1);

#if 1
  if (maxdepth >= max_depth_in_box(x, y, z, sx, sy, sz) ||
#else
  if (maxdepth == 0 ||
#endif
      sx == 1 ||
      sy == 1 ||
      sz == 1 ||
      adaptive_error(x, y, z, sx, sy, sz, t, nt) < ADAPTIVE_ERROR) {

    /* Add the tris and return. */
    for (i = 0; i < nt; i++)
      add_triangle(&t[3*i]);
  } else {
    /* Subdivide. */
    adaptive_march_cube(maxdepth+1, x     , y     , z     ,    sx/2,    sy/2,    sz/2);
    adaptive_march_cube(maxdepth+1, x+sx/2, y     , z     , sx-sx/2,    sy/2,    sz/2);
    adaptive_march_cube(maxdepth+1, x+sx/2, y+sy/2, z     , sx-sx/2, sy-sy/2,    sz/2);
    adaptive_march_cube(maxdepth+1, x     , y+sy/2, z     ,    sx/2, sy-sy/2,    sz/2);
    adaptive_march_cube(maxdepth+1, x     , y     , z+sz/2,    sx/2,    sy/2, sz-sz/2);
    adaptive_march_cube(maxdepth+1, x+sx/2, y     , z+sz/2, sx-sx/2,    sy/2, sz-sz/2);
    adaptive_march_cube(maxdepth+1, x+sx/2, y+sy/2, z+sz/2, sx-sx/2, sy-sy/2, sz-sz/2);
    adaptive_march_cube(maxdepth+1, x     , y+sy/2, z+sz/2,    sx/2, sy-sy/2, sz-sz/2);
  }
  return 0;
}




static int
common_triangle(Vertex *v1,
                Vertex *v2)
{
  int j, k;
  for (j = 0; j < v1->tris_length; j++) {
    for (k = 0; k < 3; k++) {
      if (triangles[v1->tris[j]].v[k] == v2)
        return v1->tris[j];
    }
  }
  return -1;
}

static int
find_dangling_edges(Vertex *v,
                    Vertex *result[],
                    int     result_size)
{
  int i, j, k;
  for (j = 0; j < v->tris_length; j++) {
    for (k = 0; k < 3; k++) {
      triangles[v->tris[j]].v[k]->count = 0;
    }
  }
  for (j = 0; j < v->tris_length; j++) {
    for (k = 0; k < 3; k++) {
      triangles[v->tris[j]].v[k]->count++;
    }
  }
  i = 0;
  for (j = 0; j < v->tris_length; j++) {
    for (k = 0; k < 3; k++) {
      if (triangles[v->tris[j]].v[k]->count < 2) {
        result[i++] = triangles[v->tris[j]].v[k];
        if (i >= result_size)
          return i;
      }
    }
  }
  return i;
}

static int
split_triangle(int     dir,
               int     tri,
               Vertex *base_v,
               Vertex *first_v,
               Vertex *last_v,
               Vertex *insert_v)
{
  Vertex *t[3];
  int newtri, i;

  if (dir == 1) {
    t[0] = base_v;
    t[1] = first_v;
    t[2] = insert_v;
    triangles[tri].v[0] = base_v;
    triangles[tri].v[1] = insert_v;
    triangles[tri].v[2] = last_v;
  } else /* if (dir == -1) */ {
    t[2] = base_v;
    t[1] = first_v;
    t[0] = insert_v;
    triangles[tri].v[2] = base_v;
    triangles[tri].v[1] = insert_v;
    triangles[tri].v[0] = last_v;
  }

  newtri = add_triangle(t);

  /* Fix the "tris" lists in the vertices. */
  first_v->tris_length--;
  for (i = 0; i < first_v->tris_length; i++) {
    if (first_v->tris[i] == tri)
      first_v->tris[i] = newtri;
  }
  add_triangle_to_vertex(insert_v, tri);

  return 0;
}



int
adaptive_patch_mesh()
{
  int i, j, k;
  Vertex *v[3];
  Vertex *first_v, *last_v, *prev_v, *base_v;
  int base_tri;
  int dir;

  /* Terminology:
   * I call vertices that are on grid edges on only the deeper-depth cube's side
   * "green vertices", although the below code allows them to be red, blue, or purple in
   * theory (haven't seen any in practice).
   * "Normal" vertices are called "gray" vertices.  Note that gray vertices can
   * also have dangling edges, although green vertices are guaranteed to have them. */
  for (i = 0; i < vertices_length; i++) {

    /* Some of these vertices aren't part of the current mesh. */
    if (vertices[i]->tris_length == 0) continue;

    /* Color the vertices. */
    switch ((vertices[i]->adj & 1 ? 1 : 0) +
            (vertices[i]->adj & 2 ? 1 : 0) +
            (vertices[i]->adj & 4 ? 1 : 0) +
            (vertices[i]->adj & 8 ? 1 : 0)) {
    case 0:
      vertices[i]->cr = 0.5;
      vertices[i]->cg = 0.0;
      vertices[i]->cb = 0.5;
      break;
    case 1:
      vertices[i]->cr = 1.0;
      vertices[i]->cg = 0.0;
      vertices[i]->cb = 0.0;
      break;
    case 2:
      vertices[i]->cr = 0.0;
      vertices[i]->cg = 1.0;
      vertices[i]->cb = 0.0;
      break;
    case 3:
      vertices[i]->cr = 0.0;
      vertices[i]->cg = 0.0;
      vertices[i]->cb = 1.0;
      break;
    case 4:
      break;
    }

    /* We're selecting for "green" vertices. */
    if (vertices[i]->adj == 0xF)
      continue;

    j = find_dangling_edges(vertices[i], v, 3);
    if (j > 2) {
      g_print("Oops. I don't know how to deal with green vertices with >2 dangling edges.\n");
      continue;
    } else if (j < 2) {
      g_print("Huhn?  How can a green vertex have <2 dangling edges?\n");
      continue;
    }

    if (v[0]->adj == 0xF) {
      first_v = v[0];
      last_v = v[1];
    } else if (v[1]->adj == 0xF) {
      first_v = v[1];
      last_v = v[0];
    } else {
      /* This is an "internal" green vertex among others.  Skip it and come back to it later. */
      continue;
    }
    first_v->patch_next = vertices[i];
    vertices[i]->patch_next = last_v;

    prev_v = vertices[i];
    while (last_v->adj != 0xF) {
      j = find_dangling_edges(last_v, v, 3);
      if (j > 2) {
        g_print("Oops. I don't know how to deal with green vertices with >2 dangling edges.\n");
        goto continue_outer;
      } else if (j < 2) {
        g_print("Huhn?  How can a green vertex have <2 dangling edges?\n");
        goto continue_outer;
      }
      if (prev_v == v[0]) {
        last_v->patch_next = v[1];
      } else if (prev_v == v[1]) {
        last_v->patch_next = v[0];
      } else {
        g_print("Internal error, v[0] != prev_v && v[1] != prev_v\n");
        goto continue_outer;
      }
      prev_v = last_v;
      last_v = last_v->patch_next;
    }

    if ((base_tri = common_triangle(first_v, last_v)) == -1) {
      g_print("Whoops!  first_v and last_v don't share a common triangle!\n");
      continue;
    }

    /* We now have a linked list of vertices and a base triangle to split up. */
    dir = 0;
    for (k = 0; k < 3; k++) {
      if (first_v == triangles[base_tri].v[k]) {
        if        (last_v == triangles[base_tri].v[(k+1)%3]) {
          dir = 1;
          base_v = triangles[base_tri].v[(k+2)%3];
          break;
        } else if (last_v == triangles[base_tri].v[(k+2)%3]) {
          dir = 2;
          base_v = triangles[base_tri].v[(k+1)%3];
          break;
        }
      }
    }
    if (dir == 0) {
      g_print("Uh-oh.  Internal error, dir == 0.\n");
      continue;
    }

    /* We now have the base vertex and a direction of winding of the list. */

    while (first_v->patch_next != last_v) {
      first_v->adj = 0xF;
      split_triangle(dir, base_tri, base_v, first_v, last_v, first_v->patch_next);
      first_v = first_v->patch_next;
    }
    first_v->adj = 0xF;

  continue_outer:
  }

  return 0;
}
